$ifndef _BUFFER_TX_RX_24_INC_
$define _BUFFER_TX_RX_24_INC_
'
' Interrupt driven buffered USART1 TX and RX for PIC24 and dsPIC devices
'
' Implements replacements for the compiler's HRSOut and HRSIn commands (with and without timeouts)
' With different buffer sizes for TX and RX
' Written by Les Johnson for the Proton24 compiler
'
    #Disable HRSIn, HRSInTo, HRSIn_RCREG_Read, HRSOut       ' Disable the library subroutines for HRSOut and HRSIn with and without timeouts
'
' Create the default buffer lengths for both TX and RX
'
$ifndef cUSART1_RxBufferSize    
    $define cUSART1_RxBufferSize 64                         ' RX Serial buffer size must be a power of 2. i.e. 2, 4, 8, 16, 32, 64 or 128
$endif 

$ifndef cUSART1_TxBufferSize    
    $define cUSART1_TxBufferSize 32                         ' TX Serial buffer size must be a power of 2. i.e. 2, 4, 8, 16, 32, 64 or 128
$endif 
'
' Create the variables for the interrupt driven USART1 TX and RX routines
'   
    Dim USART1_bRxIndexIn As Byte                           ' \
    Dim USART1_bRxIndexOut As Byte                          ' / USART1 RX buffer index variables    
    Dim USART1_bTxIndexIn As Byte                           ' \
    Dim USART1_bTxIndexOut As Byte                          ' / USART1 TX buffer index variables
    Dim USART1_bSerTemp As Byte                             ' Temporary variable used in the interrupt routine
    Dim USART1_bTxTemp As Byte                              ' Temporary variable used in Hrsout routine
    Dim PRTA1 As Word                                       ' Timeout value  
    
    Dim USART1_bRxBuffer[cUSART1_RxBufferSize] As Byte      ' USART1 RX circular buffer
    Dim USART1_bTxBuffer[cUSART1_TxBufferSize] As Byte      ' USART1 TX circular buffer
    
$define cUSART1_RxFIFO_Mask $eval cUSART1_RxBufferSize - 1  ' Create a mask for resetting or testing the serial RX buffer
$define cUSART1_TxFIFO_Mask $eval cUSART1_TxBufferSize - 1  ' Create a mask for resetting or testing the serial TX buffer    
$define cHrsin1_TimeoutPadding $eval ((113 * _xtal) / 2) - 1' Hrsin1 timeout delay padding 

$ifndef False
    $define False 0
$endif 
$ifndef True
    $define True 1
$endif 

$if OSENABLE_BUFFERED_SERIO = True
    Dim e_RXEvent As Word
    Dim e_RXOErr As Word
    Dim e_TXEvent As Word
    Dim RXWaitFor As Word
    $define OSNO_WAITFOR $FF00
$endif

'-------------------------------------------------------------------------------------------
' Enable the USART1 RX Interrupt                                    
'
$define USART1_Isr_RxEnable() Set IEC0bits_U1RXIE

'-------------------------------------------------------------------------------------------
' Disable the USART1 RX interrupt
'
$define USART1_Isr_RxDisable() Clear IEC0bits_U1RXIE

'-------------------------------------------------------------------------------------------
' Enable the USART1 TX Interrupt                                    
'
$define USART1_Isr_TxEnable() Set IEC0bits_U1TXIE

'-------------------------------------------------------------------------------------------
' Disable the USART1 TX interrupt
'
$define USART1_Isr_TxDisable() Clear IEC0bits_U1TXIE
    
'-------------------------------------------------------------------------------------------                                                
    GoTo _Buffer_TxRx24_Main_                               ' Jump over the subroutines                           
'-------------------------------------------------------------------------------------------   
' Service a USART1 RX interrupt
'
#ifSym __syscom_hrsin1_req_                                 ' Allow this block of code if Hrsin has been used within the main program
Isr U1RXInterrupt    
    If U1STA & $000E > 0 Then                               ' Check for any receive errors
      OSSignalSemaphore_ISR(e_RXOErr)                       ' Raise error event
      #ifSym __SYSCOM_HSERIAL1_CLEAR_REQ_                   ' Has the HSERIAL_CLEAR declare been used in the program?
        If U1STABits_OERR = 1 Then Clear U1STABits_OERR     ' Clear Overrun errors
      #endIfSym
    EndIf                                                       
    If U1StaBits_URXDA = 1 Then                             ' No errors so load all fifo data into buffer
      USART1_bRxBuffer[USART1_bRxIndexIn] = U1RXREG         ' Place the byte received into the USART1 buffer      
      USART1_bSerTemp = USART1_bRxIndexIn + 1               ' Increment the index in WREG0.Byte0 
      USART1_bSerTemp = USART1_bSerTemp & cUSART1_RxFIFO_Mask ' Reset the USART1 buffer when full
      If USART1_bSerTemp <> USART1_bRxIndexOut Then         ' Have we reached the end of the USART1 buffer?               
        USART1_bRxIndexIn = USART1_bSerTemp                 ' No. So move to the next position
      EndIf
      Clear IFS0bits_U1RXIF                                 ' Clear the interrupt request flag
    $if OSENABLE_BUFFERED_SERIO = True
      If RXWaitFor = OSNO_WAITFOR Then     
        OSSignalSemaphore_ISR(e_RXEvent)                    ' Signal when any data available                        
      Else
        If U1RXREG.LowByte = RXWaitFor.LowByte Then
          OSSignalSemaphore_ISR(e_RXEvent)                  ' Signal when Waitfor char received
          RXWaitFor = OSNO_WAITFOR                          ' clear to default no waitfor
        EndIf
      EndIf
    $endif
    EndIf    
EndIsr                                                      ' Exit the interrupt handler
#endIfSym

'------------------------------------------------------------------------------------------- 
' Service a USART1 TX interrupt
'
#ifSym __syscom_hrsout1_req_                                ' Allow this block of code if Hrsout has been used within the main program
Isr U1TXInterrupt                                      
    Clear IFS0bits_U1TXIF                                   ' Clear the TX interrupt flag
    U1TXREG = USART1_bTxBuffer[USART1_bTxIndexOut]          ' Transmit a byte from the USART1 buffer      
    Inc USART1_bTxIndexOut                                  ' Increment the out USART1 buffer Index 
    USART1_bTxIndexOut = USART1_bTxIndexOut & cUSART1_TxFIFO_Mask ' Reset when full            
    If USART1_bTxIndexOut = USART1_bTxIndexIn Then          ' Are there any more bytes in the USART1 buffer?
      USART1_Isr_TxDisable()                                ' No. So disable USART1 TX interrupts 
$if OSENABLE_BUFFERED_SERIO = True
      OSSignalSemaphore_ISR(e_TXEvent)                      ' and Signal buffer empty
$endif
    EndIf                       
EndIsr                                                      ' Exit the interrupt handler
#endIfSym

'-------------------------------------------------------------------------------------------
' Hrsin replacement code with timeout
' Input     : PRTA1 = timeout value
' Output    : WREG0.Byte0
'           : Carry flag clear if timed out
' Notes     : Follows exactly the method used by the compiler's library routine
'
#ifSym __syscom_hrsin1_to_req_                                    ' Allow this block of code if Hrsin has been used within the main program
__hrsin1__to:                                                     
    Input __HRSIN1_PORT.__HRSIN1_PIN                              ' Make the Hrsin1 pin an input (just in case it is altered elsewhere)    
    Push.d WREG2                                                  ' Save WREG2 and WREG3
    WREG2 = PRTA1                                                 ' Move the timeout value (passed from the command) to WREG2
    Clear WREG3                                                   ' Clear the inner loop counter
    Do                                                            ' Create an infinite loop
      DelayCS 1                                                   ' Padding for 1 cycle
      If USART1_bRxIndexIn <> USART1_bRxIndexOut Then             ' Is there any data in the buffer?
        Pop.d WREG2                                               ' Yes. So restore WREG2 and WREG3
        GoTo __Hrsin1__GetByte                                    ' Get the byte from USART1, and exit the routine from there
      EndIf
      Clear WREG0.Byte0                                           ' \
      Dec WREG3.Byte0                                             ' | Decrement the inner loop counter
      Subb.b WREG3H                                               ' / So that it will trigger a Carry change
      If SRbits_C = 1 Then Continue                               ' Continue if gone below zero?
      Dec WREG2.Byte0                                             ' \ Decrement the outer loop counter
      Subb.b WREG2H                                               ' / So that it will trigger a Carry change
      If SRbits_C = 0 Then                                        ' Is the carry flag clear?
        Pop.d WREG2                                               ' Yes. So restore WREG2 and WREG3
        Return                                                    ' Exit the routine
      EndIf
      WREG3 = cHrsin1_TimeoutPadding                              ' Reload the timout padding into WREG3
    Loop
#endIfSym  ' __syscom_hrsin1_to_req_ 
     
'-------------------------------------------------------------------------------------------
' Retrieve a byte from the USART1 RX buffer  
' Input     : None
' Output    : USART1_bTxTemp and WREG0.Byte0 hold the oldest byte from the buffer
' Notes     : Waits for a byte if none present (blocking)
'           : Replacement for the compiler's HRSIn command without timeout
'
#ifSym __syscom_hrsin1_req_                                         ' Allow this block of code if Hrsin has been used within the main program
__hrsin1:
    Input __HRSIN1_PORT.__HRSIN1_PIN                                ' Make the Hrsin1 pin an input (just in case it is altered elsewhere)
    Repeat                                                          ' \
    Until USART1_bRxIndexIn <> USART1_bRxIndexOut                   ' / Wait until a character is present in the buffer
'
' Get a byte from the buffer (non-blocking)
'
__Hrsin1__GetByte:    
    USART1_Isr_RxDisable()                                          ' Disable USART1 RX interrupts
    USART1_bTxTemp = USART1_bRxBuffer[USART1_bRxIndexOut]           ' Get the oldest byte received from the USART1 buffer
    Inc USART1_bRxIndexOut                                          ' Move the Index to discard USART1 buffer
    USART1_bRxIndexOut = USART1_bRxIndexOut & cUSART1_RxFIFO_Mask   ' If the Index is at end, roll-it over.
    USART1_Isr_RxEnable()                                           ' Enable USART1 RX interrupts
    WREG0.Byte0 = USART1_bTxTemp                                    ' Place the byte received into WREG0.Byte0
    Set SRbits_C                                                    ' Set the Carry flag to indicate a byte received
    Return
#endIfSym   ' __syscom_hrsin1_req_
   
'-------------------------------------------------------------------------------------------
' Transmit a single byte from the buffer to USART1
' Input     : WREG0.Byte0 holds the byte to transmit
' Output    : WREG0.Byte0 still holds the byte transmitted
' Notes     : If the buffer is full, wait until an empty slot is present
'           : Replacement for the compiler's HRSOut command
'
#ifSym __syscom_hrsout1_req_                                        ' Allow this block of code if Hrsout has been used within the main program
__hrsout1:
    Output __HRSOUT1_PORT.__HRSOUT1_PIN                             ' Make the Hrsout1 pin an output (just in case it is altered elsewhere)
    USART1_bTxTemp = WREG0.Byte0                                    ' Transfer WREG0.Byte0 to its working variable
    Do
      WREG1.Byte0 = USART1_bTxIndexIn + 1                           ' \
      WREG1.Byte0 = WREG1.Byte0 & cUSART1_TxFIFO_Mask               ' | Wait until the USART1 TX buffer has an empty slot
    Loop Until WREG1.Byte0 <> USART1_bTxIndexOut                    ' /
    USART1_Isr_TxDisable()                                          ' Disable the USART1 TX interrupt
    USART1_bTxBuffer[USART1_bTxIndexIn] = USART1_bTxTemp            ' Place a byte to transmit into the buffer
    Inc USART1_bTxIndexIn                                           ' Increment the In buffer Index
    USART1_bTxIndexIn = USART1_bTxIndexIn & cUSART1_TxFIFO_Mask     ' Reset when full 
    USART1_Isr_TxEnable()                                           ' Enable the USART1 TX Interrupt
    WREG0.Byte0 = USART1_bTxTemp                                    ' Restore WREG0.Byte0 before leaving
    Return                                                                                                              
#endIfSym  ' __syscom_hrsout1_req_, __syscom_hrsout1ln_req_


'-------------------------------------------------------------------------------------------
' Initialise the serial buffers and interrupts
' Input     : None
' Output    : None
' Notes     : None
'
Proc USART1_Init()
#ifSym __syscom_hrsin1_req_                                         ' Allow this block of code if Hrsin has been used within the main program
    USART1_Isr_RxEnable()                                           ' Enable a USART1 RX interrupt
    Clear USART1_bRxBuffer                                          ' Clear the RX buffer
    Clear USART1_bRxIndexIn                                         ' \
    Clear USART1_bRxIndexOut                                        ' / Reset the RX data in and out pointers    
    Set IPC2bits_U1RXIP0                                            ' \
    Set IPC2bits_U1RXIP1                                            ' | USART1 RX interrupt priority
    Set IPC2bits_U1RXIP2                                            ' /
    Clear IFS0bits_U1RXIF                                           ' Clear the RX interrupt request flag
$if OSENABLE_BUFFERED_SERIO = True    
    e_RXEvent = OSCreateSem(0)                                      ' Create Data Received Event
    e_RXOErr  = OSCreateBinSem(False)                               ' Create Overflow Error Event
    e_TXEvent = OSCreateBinSem(True)                                ' Create TX Empty Event and set to signalled (empty)
    RXWaitFor = OSNO_WAITFOR                                        ' Set default WaitFor value
$endif    
    
#endIfSym
#ifSym __syscom_hrsout1_req_                                        ' Allow this block of code if Hrsout has been used within the main program
    USART1_Isr_TxDisable()                                          ' Disable a USART1 TX interrupt
    Clear USART1_bTxBuffer                                          ' Clear the TX buffer
    Clear USART1_bTxIndexIn                                         ' \
    Clear USART1_bTxIndexOut                                        ' / Clear USART1 TX buffer Indexes
    Set IPC3bits_U1TXIP0                                            ' \
    Set IPC3bits_U1TXIP1                                            ' | USART1 TX interrupt priority
    Set IPC3bits_U1TXIP2                                            ' /
    Set IFS0bits_U1TXIF                                             ' Set the TX interrupt request flag so that an initial interrupt will take place
#endIfSym
EndProc

'-------------------------------------------------------------------------------------------
_Buffer_TxRx24_Main_:
$if OSENABLE_BUFFERED_SERIO = False
  #ifSym __syscom_hrsin1_req_, __syscom_hrsout1_req_
    USART1_Init()                                                   ' Initialise the interrupts etc...
  #endIfSym
$endif
    
$endif  ' _BUFFER_TX_RX_24_INC_
